
# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-08-02T15:40:06-0600",
 "dirty": true,
 "error": null,
 "full-revisionid": "1f99d246c76029c2e4e3f09f0000e382e4424e92",
 "version": "0.1.12"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
